(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translate.ml,v 1.5 2014-02-11 20:04:08 fclement Exp $ *)

module Nat = Unatural;;
module Ddgraph = Udelayed_dgraph;;

module Midx = Umatrix_index;;
module Muniv = Umatrix_universe;;
module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;

let zonation_of_partition z = Structural_mapping.of_partition z.Ddgraph.f;;

module Config = Configuration;;

open Refinement_types;;
open Zonation;;

let init_parameter_width = Nat._0_
and init_parameter_height = Nat._0_
;;

let { Uaref.get = get_parameter_width; set = set_parameter_width; } =
  Uaref.make_aref_unique "parameter width" init_parameter_width
and { Uaref.get = get_parameter_height; set = set_parameter_height; } =
  Uaref.make_aref_unique "parameter height" init_parameter_height
;;

let { Uaref.get = get_geometry_length; set = set_geometry_length; } =
  Uaref.make_aref_unique "geometry length" Nat._0_
;;

let set_size (ni, nj) =
  set_parameter_height ni;
  set_parameter_width nj;
  set_geometry_length (Nat.mul ni nj);
;;

let zone_add_index zn accu i pp_i =
  if Nat.to_int pp_i = zn then Nat.of_int i :: accu else accu
;;

let part_add_geometry mu pp zn m_zn =
  let rev_indices = Uarray.fold_left_i (zone_add_index zn) [] pp in
  let uindices = Ulist1.of_list (Ulist.rev_map Midx.of_natural rev_indices) in
  let domi = Set.of_list1 mu uindices in
  Mapping.constant domi m_zn
;;

let add_geometry mu ppm cn ma =
  let pp = Uarray_ng.matrix_get_line ppm cn in
  Array.mapi (part_add_geometry mu pp) ma
;;

let part_set_geometry ppa zn { Mapping.domain = dom; _; } =
  let zn = Nat.of_int zn in
  Set.iter (Set.array_set ppa zn) dom
;;

let get_geometry len pp =
  let ppa = Uarray.make len Nat._0_ in
  Array.iteri (part_set_geometry ppa) pp;
  ppa
;;

let set_values pp ma = Uarray.map2 Mapping.change_constant pp ma;;

let get_values pp = Array.map Mapping.get_constant pp;;

let mapping_from_array { Mapping.domain = dom; _; } = Mapping.from_array dom;;

open Scalar_spec;;
open Components_spec;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Comps : COMPONENTS) ->
struct
  module Stt = State_types.Make (Scal) (Comps);;
  
  type scalar = Scal.t;;
  type 'a components = 'a Comps.t;;
  
  type gradients = Stt.gradients;;
  type partitions = Stt.partitions;;
  type state = Stt.state;;
  type selected_cutting = Stt.selected_cutting;;
  
  (* float array array -> scal array comps *)
  let sac_of_faa faa =
    Comps.of_array (Uarray_ng.array_array_map Scal.of_float faa)
  ;;
  
  (* scal array comps -> float array array *)
  let sac_to_faa sas =
    Uarray_ng.array_array_map Scal.to_float (Comps.to_array sas)
  ;;
  
  (* Init. *)
  
  let inj_init = Config.get_dimension;;
  
  let partitions_unique_domains pps =
    let pp_0 = Comps.get pps Nat._0_ in
    Comps.map_i (partition_unique_domains pp_0) pps
  ;;
  
  let proj_coarse_init sz ppm maa j =
    set_size sz;
    let mu = Muniv.make sz in
    let mas = sac_of_faa maa in
    let pps = Comps.map_i (add_geometry mu ppm) mas in
    let pps =
      match Config.get_dimension_management () with
      | Best_component_only -> pps
      | Vector -> partitions_unique_domains pps in
    let zs = Comps.map Scal.initial_zonation pps in
    let j = Scal.of_float j in
    (zs, j)
  ;;
  
  (* Fine. *)
  
  let inj_fine ps =
    let pas = Comps.map Mapping.to_array ps in
    Umatrix.of_array_array (sac_to_faa pas)
  ;;
  
  let proj_fine ps gm =
    let gas = sac_of_faa (Umatrix.to_array_array gm) in
    Comps.map2 mapping_from_array ps gas
  ;;
  
  let make_fine (zs, _j) =
    let ps = Comps.map to_parameter zs in
    ((fun () -> inj_fine ps), proj_fine ps)
  ;;
  
  (* Coarse. *)
  
  let inj_coarse len pps =
    let ppas = Comps.map (get_geometry len) pps in
    let ppm = Umatrix.of_array_array (Comps.to_array ppas) in
    let mas = Comps.map get_values pps in
    let maa = sac_to_faa mas in
    (ppm, maa)
  ;;
  
  let proj_coarse zs pps maa j =
    let mas = sac_of_faa maa in
    let pps = Comps.map2 set_values pps mas in
    let zs = Comps.map2 zonation_of_partition zs pps in
    let j = Scal.of_float j in
    (zs, j)

  ;;
  
  let make_coarse (zs, _j) pps =
    let len = get_geometry_length () in
    ((fun () -> inj_coarse len pps), proj_coarse zs pps)
  ;;
end
;;
